from .click import common_config
from .metrics import ScalarMetricCollection
from .misc import print_with_time
from .torch import set_seed

__all__ = ["set_seed", "print_with_time", "common_config", "ScalarMetricCollection"]



import random
from typing import Union, Collection, List, Callable, Optional

import numpy as np
import torch
from torch import Tensor
from torch.optim.optimizer import Optimizer
from torch.utils.data import Sampler

from fandak.utils.misc import is_listy


def set_seed(seed: int = 1, fully_deterministic: bool = False):
    """
    Sets the python, numpy and pytorch's ransom seeds.
    But this doesn't mean that you are fully deterministic.
    If you set fully_deterministic to True, your code will run slower, more deterministic, but not fully deterministic.
    You should also care about num_workers in your code and set them to 0.
    Also see: https://pytorch.org/docs/stable/notes/randomness.html
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if fully_deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def change_multiprocess_strategy(strategy: str = "file_system"):
    """
    You might want to use this because of the following bug:
    https://github.com/pytorch/pytorch/issues/973
    """
    torch.multiprocessing.set_sharing_strategy(strategy)


def send_to_device(
    items: Collection[Tensor], device: torch.device
) -> Collection[Tensor]:
    """
    Send a single for a tuple of tensors
    """
    return tuple(map(lambda x: x.to(device), items))


def tensor_to_numpy(
    x: Union[Tensor, Collection[Tensor]]
) -> Union[np.ndarray, Collection[np.ndarray]]:
    if is_listy(x):
        return tuple(map(lambda t: tensor_to_numpy(t), x))
    else:
        return x.detach().cpu().numpy()


class OverfitSampler(Sampler):
    def __init__(self, main_source, indices):
        super().__init__(main_source)
        self.main_source = main_source
        self.indices = indices

        main_source_len = len(self.main_source)

        how_many = int(round(main_source_len / len(self.indices)))
        self.to_iter_from = []
        for _ in range(how_many):
            self.to_iter_from.extend(self.indices)

    def __iter__(self):
        return iter(self.to_iter_from)

    def __len__(self):
        return len(self.main_source)


class GeneralDataClass:
    def filter_attributes(
        self,
        the_filter: Callable[["GeneralDataClass", str], bool],
        initial_attr_list: Optional[Collection[str]] = None,
    ) -> List[str]:
        if initial_attr_list is None:
            initial_attr_list = dir(self)
        return [a for a in initial_attr_list if the_filter(self, a)]

    def get_attribute_names(self) -> List[str]:
        return self.filter_attributes(
            lambda gdc, a: not a.startswith("__") and not callable(getattr(gdc, a))
        )

    def get_tensor_attributes(self) -> List[str]:
        return self.filter_attributes(
            lambda gdc, a: not a.startswith("__")
            and not callable(getattr(gdc, a))
            and isinstance(getattr(gdc, a), torch.Tensor)
        )

    def to(self, device: torch.device):
        for attr_name in self.get_tensor_attributes():
            attr = getattr(self, attr_name)
            setattr(self, attr_name, attr.to(device))

    def item(self):
        """
        This is like calling `.item()` for all of the attributes and setting their values to it.
        Will raise error if more than a single item is inside the tensor.
        """
        for attr_name in self.get_attribute_names():
            attr = getattr(self, attr_name)
            if isinstance(attr, torch.Tensor):
                setattr(self, attr_name, attr.item())

    def pin_memory(self) -> "GeneralDataClass":
        """
        A general implementation of pin_memory. If the attribute type is Tensor, it will call pin_memory on it.
        """
        for attr_name in self.get_attribute_names():
            attr = getattr(self, attr_name)
            if isinstance(attr, torch.Tensor):
                setattr(self, attr_name, attr.pin_memory())
        return self


class MockParamGroup(dict):
    def __init__(self, values):
        super().__init__()
        self.values = values

    def __getitem__(self, item):
        assert item in ["lr", "initial_lr"]
        if item == "lr":
            return self.values[-1]
        elif item == "initial_lr":
            return self.values[0]

    def __setitem__(self, item, value):
        assert item in ["lr"]
        return self.values.append(value)


class LRVisualizer(Optimizer):
    def __init__(self, lr):
        self.the_defaults = {"lr": lr}
        self.values = [lr]

    def __getattribute__(self, attr):
        if attr == "param_groups":
            return [MockParamGroup(self.values)]
        elif attr == "_step_count":
            return 1  # just to prevent the warning.
        elif attr == "defaults":
            return self.the_defaults
        else:
            return super().__getattribute__(attr)

    def plot(self):
        import matplotlib.pylab as plt

        ax = plt.gca()
        ax.plot(self.values)



from .core.datasets import Dataset
from .core.datasets import GeneralBatch
from .core.evaluators import Evaluator
from .core.evaluators import GeneralEvaluatorResult
from .core.models import GeneralForwardOut, GeneralLoss
from .core.models import Model
from .core.trainers import Trainer

name = "fandak"
__version__ = "0.1.2.4"

__all__ = [
    "Dataset",
    "Model",
    "Trainer",
    "Evaluator",
    "GeneralBatch",
    "GeneralLoss",
    "GeneralForwardOut",
    "GeneralEvaluatorResult",
]



from collections import defaultdict
from numbers import Number
from pathlib import Path
from pickle import dump
from typing import Optional

from torch import Tensor
from torch.utils.tensorboard import SummaryWriter

from fandak.utils.misc import print_with_time
from fandak.utils.torch import GeneralDataClass


class ScalarMetricCollection:
    def __init__(
        self,
        writer: Optional[SummaryWriter],
        root: Path,
        base_name: str,
        print_each_iter: bool = False,
        report_average: bool = True,
    ):
        self.writer = writer
        self.root = root
        self.base_name = base_name
        self.print_each_iter = print_each_iter
        self.report_average = report_average
        self.values = defaultdict(list)
        self.average_base_tag = f"training_average/{self.base_name}"

    def add_value(self, dc_value: GeneralDataClass, step: int):
        def is_scalar_like(ds: GeneralDataClass, an: str) -> bool:
            a = getattr(ds, an)
            if isinstance(a, Tensor):  # is Tensor
                try:  # has .item()
                    a.item()
                except ValueError:
                    return False
                return True
            elif isinstance(a, Number):  # is Number
                return True
            else:
                return False

        loss_like_attr_names = dc_value.filter_attributes(
            is_scalar_like, initial_attr_list=dc_value.get_attribute_names()
        )

        for attr_name in loss_like_attr_names:
            tag_name = f"{self.base_name}/{attr_name}"
            attr = getattr(dc_value, attr_name)
            if isinstance(attr, Tensor):
                value = attr.item()
            else:
                value = attr
            if self.writer:
                self.writer.add_scalar(tag_name, scalar_value=value, global_step=step)
            self.values[attr_name].append(value)
            if self.print_each_iter:
                print_with_time(f"(step {step}) {tag_name}: {value}")

    def epoch_finished(self, epoch_num: int):
        if self.report_average:
            average_values = {}
            for attr_name in self.values.keys():
                average_value = self.average_value(attr_name)
                average_values[attr_name] = average_value
                tag_name = f"{self.average_base_tag}/{attr_name}"
                if self.writer:
                    self.writer.add_scalar(
                        tag=tag_name,
                        scalar_value=average_value,
                        global_step=epoch_num + 1,
                    )
                print_with_time(f"{tag_name}: {average_value}")
            self.save(
                dictionary=average_values,
                name=Path(str(epoch_num + 1)) / self.base_name,
            )
        self.reset_values()

    def reset_values(self):
        for k in self.values.keys():
            self.values[k].clear()

    def average_value(self, attr_name: str) -> float:
        return sum(self.values[attr_name]) / len(self.values[attr_name])

    def save(self, dictionary: dict = None, name: str = None):
        if dictionary is None:
            dictionary = self.values
        if name is None:
            name = f"{self.base_name}"
        file_path = self.root / f"{name}.pkl"
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(file_path, "wb") as f:
            dump(dictionary, f)



import click


def common_config(function):
    function = click.option(
        "--cfg",
        "file_configs",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
        multiple=True,
    )(function)
    function = click.option("--set", "set_configs", type=(str, str), multiple=True)(
        function
    )

    return function



from typing import List

from yacs.config import CfgNode


def update_config(
    default_config: CfgNode,
    file_configs: List[str],
    set_configs: List[str],
    freeze: bool = True,
) -> CfgNode:
    """
    This is useful for updating your config from CLI inputs, whether by new config files or by --set.
    It will freeze the default config to prevent code smell.
    """
    cfg = default_config
    # updating config from file
    for fc in file_configs:
        cfg.merge_from_file(fc)
    # updating config from set
    for sc in set_configs:
        cfg.merge_from_list(list(sc))

    if freeze:
        cfg.freeze()

    return cfg
